use std::{
    fmt,
    sync::mpsc,
    thread,
    time::Duration,
};

use rayon::ThreadPool;
use thiserror::Error;

fn main() 
{
    let pool = rayon::ThreadPoolBuilder::new ().build ().unwrap ();
    let mut driver = Driver::new (&pool);
    driver.main ();
}

struct Driver <'a>
{
    pool: &'a ThreadPool,
    send: mpsc::SyncSender <TaskOutput>,
    recv: mpsc::Receiver <TaskOutput>,
    
    ctl: Controller,
    
    capture: Option <Capture>,
    encoder: EncoderHandle,
    transmitter: Option <Transmitter>,
}

impl <'a> Driver <'_>
{
    fn new (pool: &'a ThreadPool) -> Driver <'a>
    {
        let (send, recv) = mpsc::sync_channel (8);
        
        Driver
        {
            pool,
            send,
            recv,
            
            ctl: Default::default (),
            
            capture: Some (Default::default ()),
            encoder: EncoderHandle::Stopped (Default::default ()),
            transmitter: Some (Default::default ()),
        }
    }
    
    fn main (&mut self)
    {
        for _ in 0..10
        {
            let ev = self.ctl.poll ();
            
            if let Some (event) = ev.tx
            {
                dbg! (&event);
                self.handle_tx_event (event);
                
                match self.recv.try_recv ()
                {
                    Ok (output) => self.handle_task_output (output),
                    Err (_) => (),
                }
            }
            else
            {
                match self.recv.recv ()
                {
                    Ok (output) => self.handle_task_output (output),
                    Err (_) => (),
                }
            }
        }
    }
    
    fn handle_task_output (&mut self, output: TaskOutput)
    {
        dbg! (&output);
        match output
        {
            TaskOutput::TxCapture ((cap, buf_raw)) =>
            {
                let _old = match std::mem::replace (&mut self.capture, Some (cap))
                {
                    None => (),
                    _ => panic! ("tried to finish an already finished capture"),
                };
                
                self.ctl.tx_pipe.handle_capture_frame (buf_raw).unwrap ();
            },
            TaskOutput::TxEncode (enc) =>
            {
                let _metadata = match std::mem::replace (&mut self.encoder, EncoderHandle::Stopped (enc))
                {
                    EncoderHandle::Running (_) => (),
                    _ => panic! ("tried to finish an already finished encode"),
                };
                
                self.ctl.tx_pipe.encoder_is_busy = false;
            },
            TaskOutput::TxTransmit ((tx, busy)) =>
            {
                let _old = match std::mem::replace (&mut self.transmitter, Some (tx))
                {
                    None => (),
                    _ => panic! ("tried to finish an already finished transmit"),
                };
                
                self.ctl.tx_pipe.handle_network_busy (busy);
            }
        }
    }
    
    fn handle_tx_event (&mut self, event: TxPipelineEvent)
    {
        let dur_capture = Duration::from_millis (30);
        let dur_encode = Duration::from_millis (20);
        let dur_transmit = Duration::from_millis (200);
        
        let send = self.send.clone ();
        
        match event
        {
            TxPipelineEvent::Capture => {
                let cap = match std::mem::replace (&mut self.capture, None)
                {
                    Some (x) => x,
                    _ => {
                        dbg! ("capture is already running");
                        return;
                    },
                };
                
                self.pool.spawn (move || 
                {
                    thread::sleep (dur_capture);
                    let buf_raw = BufRaw
                    {
                        buf: vec![],
                    };
                    send.send (TaskOutput::TxCapture ((cap, buf_raw))).unwrap ();
                });
            },
            TxPipelineEvent::PollEncoder =>
            {
                let encoder = match &mut self.encoder
                {
                    EncoderHandle::Running (_) => panic! ("tried to poll encoder while it was running"),
                    EncoderHandle::Stopped (x) => x,
                };
                
                let mut buf = vec! [];
                
                let may_have_data = encoder.poll_encoded (&mut buf);
                if may_have_data
                {
                    self.ctl.tx_pipe.handle_encoded_packet (BufEncoded {buf}).unwrap ();
                }
                self.ctl.tx_pipe.encoder_has_data = may_have_data;
            }
            TxPipelineEvent::Encode (buf_raw) => 
            {
                let metadata = EncoderTaskMetadata {};
                let mut encoder = match std::mem::replace (&mut self.encoder, EncoderHandle::Running (metadata))
                {
                    EncoderHandle::Stopped (x) => x,
                    _ => panic! ("tried to run the same encoder twice"),
                };
                
                self.pool.spawn (move || 
                {
                    thread::sleep (dur_encode);
                    encoder.try_handle_insert_data (&buf_raw.buf).unwrap ();
                    send.send (TaskOutput::TxEncode (encoder)).unwrap ();
                });
            },
            TxPipelineEvent::Transmit (_buf_enc) => {
                let tx = match std::mem::replace (&mut self.transmitter, None)
                {
                    Some (x) => x,
                    _ => panic! ("tried to run the same transmitter twice"),
                };
                
                self.pool.spawn (move ||
                {
                    thread::sleep (dur_transmit);
                    send.send (TaskOutput::TxTransmit ((tx, false))).unwrap ();
                });
            },
        }
    }
}

enum EncoderHandle
{
    Stopped (Encoder),
    Running (EncoderTaskMetadata),
}

struct EncoderTaskMetadata
{
    
}

enum TaskOutput
{
    TxCapture ((Capture, BufRaw)),
    TxEncode (Encoder),
    TxTransmit ((Transmitter, bool)),
}

impl fmt::Debug for TaskOutput {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self
        {
            Self::TxCapture (_) => f.debug_tuple ("TaskOutput::TxCapture").finish (),
            Self::TxEncode (_) => f.debug_tuple ("TaskOutput::TxEncode").finish (),
            Self::TxTransmit (_) => f.debug_tuple ("TaskOutput::TxTransmit").finish (),
        }
    }
}

#[derive (Default)]
struct Controller
{
    /// Pipeline for pulling data from the network, decoding it, 
    /// and displaying it.
    
    rx_pipe: RxPipeline,
    
    /// Pipeline for pulling data from capture devices, encoding it,
    /// and sending it over the network.
    
    tx_pipe: TxPipeline,
}

#[derive (Debug)]
struct ControlEvent
{
    rx: Option <RxPipelineEvent>,
    tx: Option <TxPipelineEvent>,
}

impl Controller
{
    fn poll (&mut self) -> ControlEvent
    {
        let rx = self.rx_pipe.poll ();
        let tx = self.tx_pipe.poll ();
        
        ControlEvent {
            rx,
            tx,
        }
    }
}

#[derive (Default)]
struct RxPipeline
{
    buf_enc: Option <BufEncoded>,
    buf_raw: Option <BufRaw>,
}

#[derive (Debug)]
enum RxPipelineEvent
{
    DisplayRaw (BufRaw),
    Decode (BufEncoded),
    Receive,
}

#[derive (Debug, Error)]
enum RxPipeError
{
    #[error ("already have encoded data")]
    AlreadyHaveEncodedData,
    #[error ("already have raw data")]
    AlreadyHaveRawData,
}

impl RxPipeline
{
    fn handle_encoded_packet (&mut self, buf_enc: BufEncoded) -> Result <(), RxPipeError>
    {
        if self.has_encoded_data ()
        {
            return Err (RxPipeError::AlreadyHaveEncodedData);
        }
        
        self.buf_enc = Some (buf_enc);
        
        Ok (())
    }
    
    fn handle_raw_frame (&mut self, buf_raw: BufRaw) -> Result <(), RxPipeError>
    {
        if self.has_raw_data ()
        {
            return Err (RxPipeError::AlreadyHaveRawData);
        }
        
        self.buf_raw = Some (buf_raw);
        
        Ok (())
    }
    
    fn poll (&mut self) -> Option <RxPipelineEvent>
    {
        if let Some (buf_raw) = self.buf_raw.take ()
        {
            return Some (RxPipelineEvent::DisplayRaw (buf_raw));
        }
        
        if ! self.has_raw_data ()
        {
            if let Some (buf_enc) = self.buf_enc.take ()
            {
                return Some (RxPipelineEvent::Decode (buf_enc));
            }
        }
        
        if ! self.has_encoded_data ()
        {
            return Some (RxPipelineEvent::Receive);
        }
        
        None
    }
    
    /// True if we have a buffer of received data, ready
    /// to be decoded.
    
    fn has_encoded_data (&self) -> bool
    {
        self.buf_enc.is_some ()
    }
    
    fn has_raw_data (&self) -> bool
    {
        self.buf_raw.is_some ()
    }
}

#[derive (Default)]
struct TxPipeline
{
    buf_enc: Option <BufEncoded>,
    buf_raw: Option <BufRaw>,
    capture_is_busy: bool,
    encoder_has_data: bool,
    encoder_is_busy: bool,
    network_busy: bool,
}

#[derive (Debug)]
enum TxPipelineEvent
{
    /// Capture a raw frame and pass it to handle_capture_frame
    Capture,
    
    /// Poll the encoder. If it returns a packet, pass it to 
    /// handle_encoded_packet.
    PollEncoder,
    
    /// Encode this raw frame and pass the encoded frame to
    /// handle_encoded_packet
    Encode (BufRaw),
    
    /// Transmit this encoded frame to the network
    Transmit (BufEncoded),
}

impl TxPipeline
{
    fn handle_capture_frame (&mut self, buf_raw: BufRaw) 
    -> Result <(), ControlError>
    {
        self.capture_is_busy = false;
        
        if self.has_raw_frame ()
        {
            return Err (ControlError::AlreadyHaveRawData);
        }
        
        self.buf_raw = Some (buf_raw);
        
        Ok (())
    }
    
    fn handle_encoded_packet (&mut self, buf_enc: BufEncoded) -> Result <(), ControlError>
    {
        self.encoder_is_busy = false;
        
        if self.has_encoded_packet ()
        {
            return Err (ControlError::AlreadyHaveEncodedData);
        }
        
        self.buf_enc = Some (buf_enc);
        
        Ok (())
    }
    
    fn handle_network_busy (&mut self, busy: bool)
    {
        self.network_busy = busy;
    }
    
    fn has_encoded_packet (&self) -> bool
    {
        self.buf_enc.is_some ()
    }
    
    fn has_raw_frame (&self) -> bool
    {
        self.buf_raw.is_some ()
    }
    
    fn poll (&mut self) -> Option <TxPipelineEvent>
    {
        if ! self.network_busy
        {
            if let Some (buf_enc) = self.buf_enc.take ()
            {
                return Some (TxPipelineEvent::Transmit (buf_enc));
            }
        }
        
        if ! self.has_encoded_packet () && ! self.encoder_is_busy
        {
            if self.encoder_has_data
            {
                return Some (TxPipelineEvent::PollEncoder);
            }
            
            if let Some (buf_raw) = self.buf_raw.take ()
            {
                self.encoder_has_data = true;
                self.encoder_is_busy = true;
                return Some (TxPipelineEvent::Encode (buf_raw));
            }
        }
        
        if ! self.has_raw_frame ()
        {
            if ! self.capture_is_busy
            {
                self.capture_is_busy = true;
                return Some (TxPipelineEvent::Capture);
            }
        }
        
        None
    }
}

/// Wraps non-thread-safe capture device like a v4l webcam

#[derive (Default)]
struct Capture
{
    
}

impl Capture
{
    /// Blocks until the capture device gets us a frame
    
    fn wait_for_frame (&mut self, _buf: &mut [u8]) -> bool
    {
        false
    }
}

/// Wraps non-thread-safe encoder state like from ffmpeg

#[derive (Default)]
struct Encoder
{
    /// AVCodecContext
    ctx: (),
    
    /// Encoded AVPacket
    pkt_enc: Option <()>,
    
    /// Bogus debugging thing
    internal_buffer: usize,
}

impl Encoder
{
    /// If any encoded bytes are ready, and the caller's buffer 
    /// is big enough, copy encoded bytes into the caller's buffer, 
    /// consuming them from internal buffers.
    /// 
    /// Won't block
    
    fn poll_encoded (&mut self, _buf_enc: &mut [u8]) -> bool
    {
        if let Some (_pkt) = self.pkt_enc.take ()
        {
            // Do copy
            return true;
        }
        
        if self.internal_buffer >= 2
        {
            self.internal_buffer -= 2;
            // Do copy
            return true;
        }
        
        false
    }
    
    /// Insert raw data. If the internal buffers are full, the new
    /// data is rejected.
    /// Call poll_encoded before calling this, to drain internal
    /// buffers if needed.
    /// 
    /// Blocks if the internals perform encoding
    
    fn try_handle_insert_data (&mut self, _buf_raw: &[u8]) -> Result <(), ControlError>
    {
        if self.pkt_enc.is_some ()
        {
            return Err (ControlError::AlreadyHaveRawData);
        }
        
        if self.internal_buffer >= 3
        {
            return Err (ControlError::AlreadyHaveRawData);
        }
        
        self.internal_buffer += 3;
        
        Ok (())
    }
}

/// Wraps a network transmitter like a TCP send stream or a QUIC
/// outgoing uni stream

#[derive (Default)]
struct Transmitter
{
    
}

impl Transmitter
{
    /// Tries to transmit data over the network.
    /// If Transmitter sees congestion, it may reject the transmission
    /// until a later time.
    
    fn try_send (&mut self, _now: std::time::Instant, _buf: &[u8]) -> bool
    {
        false
    }
}

struct BufEncoded
{
    buf: Vec <u8>
}

impl fmt::Debug for BufEncoded {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f
        .debug_struct("BufEncoded")
        .finish()
    }
}

struct BufRaw
{
    buf: Vec <u8>
}

impl fmt::Debug for BufRaw {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f
        .debug_struct("BufRaw")
        .finish()
    }
}

#[derive (Debug, Error)]
enum ControlError
{
    #[error ("already have encoded data")]
    AlreadyHaveEncodedData,
    #[error ("already have raw data")]
    AlreadyHaveRawData,
}
